@@ -1,9 +1,9 @@
 /*
- * Copyright (C) 2016-2019 Red Hat, Inc.  All rights reserved.
+ * Copyright (C) 2016-2020 Red Hat, Inc.  All rights reserved.
  *
  * Authors: Fabio M. Di Nitto <fabbione@kronosnet.org>
  *
- * This software licensed under GPL-2.0+, LGPL-2.0+
+ * This software licensed under GPL-2.0+
  */
 
 #include "config.h"
@@ -46,6 +46,8 @@ static int wait_for_perf_rx = 0;
 static char *compresscfg = NULL;
 static char *cryptocfg = NULL;
 static int machine_output = 0;
+static int use_access_lists = 0;
+static int use_pckt_verification = 0;
 
 static int bench_shutdown_in_progress = 0;
 static pthread_mutex_t shutdown_mutex = PTHREAD_MUTEX_INITIALIZER;
@@ -66,6 +68,8 @@ static int test_type = TEST_PING;
 static uint64_t perf_by_size_size = 1 * ONE_GIGABYTE;
 static uint64_t perf_by_time_secs = 10;
 
+static uint32_t force_packet_size = 0;
+
 struct node {
 	int nodeid;
 	int links;
@@ -73,11 +77,17 @@ struct node {
 	struct sockaddr_storage address[KNET_MAX_LINK];
 };
 
+struct pckt_ver {
+	uint32_t len;
+	uint32_t chksum;
+};
+
 static void print_help(void)
 {
 	printf("knet_bench usage:\n");
 	printf(" -h                                        print this help (no really)\n");
 	printf(" -d                                        enable debug logs (default INFO)\n");
+	printf(" -f                                        enable use of access lists (default: off)\n");
 	printf(" -c [implementation]:[crypto]:[hashing]    crypto configuration. (default disabled)\n");
 	printf("                                           Example: -c nss:aes128:sha1\n");
 	printf(" -z [implementation]:[level]:[threshold]   compress configuration. (default disabled)\n");
@@ -107,11 +117,13 @@ static void print_help(void)
 	printf(" -s                                        nodeid that will generate traffic for benchmarks\n");
 	printf(" -S [size|seconds]                         when used in combination with -T perf-by-size it indicates how many GB of traffic to generate for the test. (default: 1GB)\n");
 	printf("                                           when used in combination with -T perf-by-time it indicates how many Seconds of traffic to generate for the test. (default: 10 seconds)\n");
+	printf(" -x                                        force packet size for perf-by-time or perf-by-size\n");
 	printf(" -C                                        repeat the test continously (default: off)\n");
 	printf(" -X[XX]                                    show stats at the end of the run (default: 1)\n");
 	printf("                                           1: show handle stats, 2: show summary link stats\n");
 	printf("                                           3: show detailed link stats\n");
 	printf(" -a                                        enable machine parsable output (default: off).\n");
+	printf(" -v                                        enable packet verification for performance tests (default: off).\n");
 }
 
 static void parse_nodes(char *nodesinfo[MAX_NODES], int onidx, int port, struct node nodes[MAX_NODES], int *thisidx)
@@ -248,7 +260,7 @@ static void setup_knet(int argc, char *argv[])
 
 	memset(nodes, 0, sizeof(nodes));
 
-	while ((rv = getopt(argc, argv, "aCT:S:s:ldom:wb:t:n:c:p:X::P:z:h")) != EOF) {
+	while ((rv = getopt(argc, argv, "aCT:S:s:lvdfom:wb:t:n:c:p:x:X::P:z:h")) != EOF) {
 		switch(rv) {
 			case 'h':
 				print_help();
@@ -260,6 +272,9 @@ static void setup_knet(int argc, char *argv[])
 			case 'd':
 				debug = KNET_LOG_DEBUG;
 				break;
+			case 'f':
+				use_access_lists = 1;
+				break;
 			case 'c':
 				if (cryptocfg) {
 					printf("Error: -c can only be specified once\n");
@@ -272,22 +287,24 @@ static void setup_knet(int argc, char *argv[])
 					printf("Error: -p can only be specified once\n");
 					exit(FAIL);
 				}
-				policystr = optarg;
-				if (!strcmp(policystr, "active")) {
-					policy = KNET_LINK_POLICY_ACTIVE;
-					policyfound = 1;
-				}
-				/*
-				 * we can't use rr because clangs can't compile
-				 * an array of 3 strings, one of which is 2 bytes long
-				 */
-				if (!strcmp(policystr, "round-robin")) {
-					policy = KNET_LINK_POLICY_RR;
-					policyfound = 1;
-				}
-				if (!strcmp(policystr, "passive")) {
-					policy = KNET_LINK_POLICY_PASSIVE;
-					policyfound = 1;
+				if (optarg) {
+					policystr = optarg;
+					if (!strcmp(policystr, "active")) {
+						policy = KNET_LINK_POLICY_ACTIVE;
+						policyfound = 1;
+					}
+					/*
+					 * we can't use rr because clangs can't compile
+					 * an array of 3 strings, one of which is 2 bytes long
+					 */
+					if (!strcmp(policystr, "round-robin")) {
+						policy = KNET_LINK_POLICY_RR;
+						policyfound = 1;
+					}
+					if (!strcmp(policystr, "passive")) {
+						policy = KNET_LINK_POLICY_PASSIVE;
+						policyfound = 1;
+					}
 				}
 				if (!policyfound) {
 					printf("Error: invalid policy %s specified. -p accepts active|passive|rr\n", policystr);
@@ -299,14 +316,16 @@ static void setup_knet(int argc, char *argv[])
 					printf("Error: -P can only be specified once\n");
 					exit(FAIL);
 				}
-				protostr = optarg;
-				if (!strcmp(protostr, "UDP")) {
-					protocol = KNET_TRANSPORT_UDP;
-					protofound = 1;
-				}
-				if (!strcmp(protostr, "SCTP")) {
-					protocol = KNET_TRANSPORT_SCTP;
-					protofound = 1;
+				if (optarg) {
+					protostr = optarg;
+					if (!strcmp(protostr, "UDP")) {
+						protocol = KNET_TRANSPORT_UDP;
+						protofound = 1;
+					}
+					if (!strcmp(protostr, "SCTP")) {
+						protocol = KNET_TRANSPORT_SCTP;
+						protofound = 1;
+					}
 				}
 				if (!protofound) {
 					printf("Error: invalid protocol %s specified. -P accepts udp|sctp\n", policystr);
@@ -375,23 +394,38 @@ static void setup_knet(int argc, char *argv[])
 				}
 				break;
 			case 'T':
-				if (!strcmp("ping", optarg)) {
-					test_type = TEST_PING;
-				}
-				if (!strcmp("ping_data", optarg)) {
-					test_type = TEST_PING_AND_DATA;
-				}
-				if (!strcmp("perf-by-size", optarg)) {
-					test_type = TEST_PERF_BY_SIZE;
-				}
-				if (!strcmp("perf-by-time", optarg)) {
-					test_type = TEST_PERF_BY_TIME;
+				if (optarg) {
+					if (!strcmp("ping", optarg)) {
+						test_type = TEST_PING;
+					}
+					if (!strcmp("ping_data", optarg)) {
+						test_type = TEST_PING_AND_DATA;
+					}
+					if (!strcmp("perf-by-size", optarg)) {
+						test_type = TEST_PERF_BY_SIZE;
+					}
+					if (!strcmp("perf-by-time", optarg)) {
+						test_type = TEST_PERF_BY_TIME;
+					}
+				} else {
+					printf("Error: -T requires an option\n");
+					exit(FAIL);
 				}
 				break;
 			case 'S':
 				perf_by_size_size = (uint64_t)atoi(optarg) * ONE_GIGABYTE;
 				perf_by_time_secs = (uint64_t)atoi(optarg);
 				break;
+			case 'x':
+				force_packet_size = (uint32_t)atoi(optarg);
+				if ((force_packet_size < 64) || (force_packet_size > 65536)) {
+					printf("Unsupported packet size %u (accepted 64 - 65536)\n", force_packet_size);
+					exit(FAIL);
+				}
+				break;
+			case 'v':
+				use_pckt_verification = 1;
+				break;
 			case 'C':
 				continous = 1;
 				break;
@@ -456,6 +490,11 @@ static void setup_knet(int argc, char *argv[])
 		exit(FAIL);
 	}
 
+	if (knet_handle_enable_access_lists(knet_h, use_access_lists) < 0) {
+		printf("Unable to knet_handle_enable_access_lists: %s\n", strerror(errno));
+		exit(FAIL);
+	}
+
 	if (cryptocfg) {
 		memset(&knet_handle_crypto_cfg, 0, sizeof(knet_handle_crypto_cfg));
 		cryptomodel = strtok(cryptocfg, ":");
@@ -625,6 +664,24 @@ static void setup_knet(int argc, char *argv[])
 	}
 }
 
+/*
+ * calculate weak chksum (stole from corosync for debugging purposes)
+ */
+static uint32_t compute_chsum(const unsigned char *data, uint32_t data_len)
+{
+	unsigned int i;
+	unsigned int checksum = 0;
+
+	for (i = 0; i < data_len; i++) {
+		if (checksum & 1) {
+			checksum |= 0x10000;
+		}
+
+		checksum = ((checksum >> 1) + (unsigned char)data[i]) & 0xffff;
+	}
+	return (checksum);
+}
+
 static void *_rx_thread(void *args)
 {
 	int rx_epoll;
@@ -737,6 +794,20 @@ static void *_rx_thread(void *args)
 							}
 							continue;
 						}
+						if (use_pckt_verification) {
+							struct pckt_ver *recv_pckt = (struct pckt_ver *)msg[i].msg_hdr.msg_iov->iov_base;
+							uint32_t chksum;
+
+							if (msg[i].msg_len != recv_pckt->len) {
+								printf("Wrong packet len received: %u expected: %u!\n", msg[i].msg_len, recv_pckt->len);
+								exit(FAIL);
+							}
+							chksum = compute_chsum((const unsigned char *)msg[i].msg_hdr.msg_iov->iov_base + sizeof(struct pckt_ver), msg[i].msg_len - sizeof(struct pckt_ver));
+							if (recv_pckt->chksum != chksum){
+								printf("Wrong packet checksum received: %u expected: %u!\n", recv_pckt->chksum, chksum);
+								exit(FAIL);
+							}
+						}
 						rx_pkts++;
 						rx_bytes = rx_bytes + msg[i].msg_len;
 						current_pckt_size = msg[i].msg_len;
@@ -855,7 +926,7 @@ static int setup_send_buffers_common(struct knet_mmsghdr *msg, struct iovec *iov
 			printf("TXT: Unable to malloc!\n");
 			return -1;
 		}
-		memset(tx_buf[i], 0, KNET_MAX_PACKET_SIZE);
+		memset(tx_buf[i], i, KNET_MAX_PACKET_SIZE);
 		iov_out[i].iov_base = (void *)tx_buf[i];
 		memset(&msg[i].msg_hdr, 0, sizeof(struct msghdr));
 		msg[i].msg_hdr.msg_iov = &iov_out[i];
@@ -879,8 +950,16 @@ static void send_perf_data_by_size(void)
 	setup_send_buffers_common(msg, iov_out, tx_buf);
 
 	while (packetsize <= KNET_MAX_PACKET_SIZE) {
+		if (force_packet_size) {
+			packetsize = force_packet_size;
+		}
 		for (i = 0; i < PCKT_FRAG_MAX; i++) {
 			iov_out[i].iov_len = packetsize;
+			if (use_pckt_verification) {
+				struct pckt_ver *tx_pckt = (struct pckt_ver *)&iov_out[i].iov_base;
+				tx_pckt->len = iov_out[i].iov_len;
+				tx_pckt->chksum = compute_chsum((const unsigned char *)iov_out[i].iov_base + sizeof(struct pckt_ver), iov_out[i].iov_len - sizeof(struct pckt_ver));
+			}
 		}
 
 		total_pkts_to_tx = perf_by_size_size / packetsize;
@@ -907,7 +986,7 @@ static void send_perf_data_by_size(void)
 
 		knet_send(knet_h, ctrl_message, TEST_STOP, channel);
 
-		if (packetsize == KNET_MAX_PACKET_SIZE) {
+		if ((packetsize == KNET_MAX_PACKET_SIZE) || (force_packet_size)) {
 			break;
 		}
 
@@ -947,15 +1026,14 @@ static void display_stats(int level)
 	struct knet_link_stats total_link_stats;
 	knet_node_id_t host_list[KNET_MAX_HOST];
 	uint8_t link_list[KNET_MAX_LINK];
-	int res;
 	unsigned int i,j;
 	size_t num_hosts, num_links;
 
-	res = knet_handle_get_stats(knet_h, &handle_stats, sizeof(handle_stats));
-	if (res) {
+	if (knet_handle_get_stats(knet_h, &handle_stats, sizeof(handle_stats)) < 0) {
 		perror("[info]: failed to get knet handle stats");
 		return;
 	}
+
 	if (compresscfg || cryptocfg) {
 		printf("\n");
 		printf("[stat]: handle stats\n");
@@ -995,8 +1073,7 @@ static void display_stats(int level)
 
 	memset(&total_link_stats, 0, sizeof(struct knet_link_stats));
 
-	res = knet_host_get_host_list(knet_h, host_list, &num_hosts);
-	if (res) {
+	if (knet_host_get_host_list(knet_h, host_list, &num_hosts) < 0) {
 		perror("[info]: cannot get host list for stats");
 		return;
 	}
@@ -1005,18 +1082,16 @@ static void display_stats(int level)
 	qsort(host_list, num_hosts, sizeof(uint16_t), node_compare);
 
 	for (j=0; j<num_hosts; j++) {
-		res = knet_link_get_link_list(knet_h, host_list[j], link_list, &num_links);
-		if (res) {
+		if (knet_link_get_link_list(knet_h, host_list[j], link_list, &num_links) < 0) {
 			perror("[info]: cannot get link list for stats");
 			return;
 		}
 
 		for (i=0; i < num_links; i++) {
-			res = knet_link_get_status(knet_h,
-						   host_list[j],
-						   link_list[i],
-						   &link_status,
-						   sizeof(link_status));
+			if (knet_link_get_status(knet_h, host_list[j], link_list[i], &link_status, sizeof(link_status)) < 0) {
+				perror("[info]: cannot get link status");
+				return;
+			}
 
 			total_link_stats.tx_data_packets += link_status.stats.tx_data_packets;
 			total_link_stats.rx_data_packets += link_status.stats.rx_data_packets;
@@ -1160,8 +1235,16 @@ static void send_perf_data_by_time(void)
 	memset(&clock_end, 0, sizeof(clock_start));
 
 	while (packetsize <= KNET_MAX_PACKET_SIZE) {
+		if (force_packet_size) {
+			packetsize = force_packet_size;
+		}
 		for (i = 0; i < PCKT_FRAG_MAX; i++) {
 			iov_out[i].iov_len = packetsize;
+			if (use_pckt_verification) {
+				struct pckt_ver *tx_pckt = (struct pckt_ver *)iov_out[i].iov_base;
+				tx_pckt->len = iov_out[i].iov_len;
+				tx_pckt->chksum = compute_chsum((const unsigned char *)iov_out[i].iov_base + sizeof(struct pckt_ver), iov_out[i].iov_len - sizeof(struct pckt_ver));
+			}
 		}
 		printf("[info]: testing with %u bytes packet size for %" PRIu64 " seconds.\n", packetsize, perf_by_time_secs);
 
@@ -1190,7 +1273,7 @@ static void send_perf_data_by_time(void)
 
 		knet_send(knet_h, ctrl_message, TEST_STOP, channel);
 
-		if (packetsize == KNET_MAX_PACKET_SIZE) {
+		if ((packetsize == KNET_MAX_PACKET_SIZE) || (force_packet_size)) {
 			break;
 		}
 
